/*
 * linux/drivers/pcmcia/ep93xx_armadillo9.c
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/device.h>
#include <linux/delay.h>

#include <pcmcia/ss.h>

#include <asm/io.h>
#include <asm/hardware.h>
#include <asm/irq.h>

#include "soc_common.h"

//
// PCMCIA status signals = GPIO port F.
// 
#define EP93XX_PCMCIA_WP          		0x01
#define EP93XX_PCMCIA_CD1         		0x02
#define EP93XX_PCMCIA_CD2         		0x04
#define EP93XX_PCMCIA_DETECT      		0x06
#define EP93XX_PCMCIA_BVD1        		0x08
#define EP93XX_PCMCIA_BVD2        		0x10
#define EP93XX_PCMCIA_VS1         		0x20
#define EP93XX_PCMCIA_RDY         		0x40
#define EP93XX_PCMCIA_VS2			0x80

#define EP93XX_PCMCIA_INT_WP			IRQ_GPIO0	/* 19 */
#define EP93XX_PCMCIA_INT_CD1			IRQ_GPIO1	/* 20 */
#define EP93XX_PCMCIA_INT_CD2			IRQ_GPIO2	/* 21 */
#define EP93XX_PCMCIA_INT_BVD1			IRQ_GPIO3	/* 22 */
#define EP93XX_PCMCIA_INT_BVD2			IRQ_GPIO4	/* 47 */
#define EP93XX_PCMCIA_INT_VS1			IRQ_GPIO5	/* 48 */
#define EP93XX_PCMCIA_INT_RDY			IRQ_GPIO6	/* 49 */
#define EP93XX_PCMCIA_INT_VS2			IRQ_GPIO7	/* 50 */

//
// Definitions to program the TPS2202A PCMCIA controller.
//
#define AVCC_0V         			0x000
#define AVCC_33V        			0x004
#define AVCC_5V         			0x008
#define BVCC_0V         			0x000
#define BVCC_33V        			0x080
#define BVCC_5V         			0x040
#define AVPP_0V         			0x000
#define BVPP_0V         			0x000
#define ENABLE          			0x100
#define EE_ADDRESS      			(0x55<<9)

static struct pcmcia_irqs irqs[] = {
	{ 0, EP93XX_PCMCIA_INT_CD1, "CF CD1" },
	{ 0, EP93XX_PCMCIA_INT_CD2, "CF CD2" },
	{ 0, EP93XX_PCMCIA_INT_BVD1, "CF STSCHG/BVD1" },
	{ 0, EP93XX_PCMCIA_INT_BVD2, "CF SPKR/BVD2" },
};

static int armadillo9_pcmcia_hw_init(struct soc_pcmcia_socket *skt)
{
	int rtn;

	if (skt->nr != 0)
		return -ENXIO;

	skt->irq = EP93XX_PCMCIA_INT_RDY;

	rtn = soc_pcmcia_request_irqs(skt, irqs, ARRAY_SIZE(irqs));

	/* enable GPIO6 as interrupt */
	writel(0x40, GPIO_FEOI);

	/* enable GPIO1-4 and 6 as interrupt */
	/* Note: 1-4 need to be edge triggered, but
	   then the irq vectors need to hit GPIOxEIO
	   which they don't currently do - no big deal
	   since CD and BVD are polled anyway */
	/* writel(0x5e, GPIO_FEOI); */

	return rtn;
}

static void armadillo9_pcmcia_hw_shutdown(struct soc_pcmcia_socket *skt)
{
	soc_pcmcia_free_irqs(skt, irqs, ARRAY_SIZE(irqs));
}

static void armadillo9_pcmcia_socket_state(struct soc_pcmcia_socket *skt,
				    struct pcmcia_state *state)
{
	unsigned long status;

	status = inl(GPIO_PFDR);

	state->detect = (status & EP93XX_PCMCIA_DETECT) ? 0 : 1;
	state->ready  = (status & EP93XX_PCMCIA_RDY)  ? 1 : 0;
	state->bvd1   = (status & EP93XX_PCMCIA_BVD1)   ? 1 : 0;
	state->bvd2   = (status & EP93XX_PCMCIA_BVD2)   ? 1 : 0;
	state->vs_3v  = (status & EP93XX_PCMCIA_VS1)  ? 0 : 1;
	state->vs_Xv  = (status & EP93XX_PCMCIA_VS2)  ? 0 : 1;
	state->wrprot = (status & EP93XX_PCMCIA_WP) ? 0 : 1;

//	printk("%s: nr: %d, status: %#x, state: %#x\n", __FUNCTION__, skt->nr, status, *((unsigned char *) state));
}

static int ep93xx_set_voltage( u_char NewVcc )
{
	static u_char Vcc = 0;
	unsigned long ulSwitchSettings;

	if( Vcc == NewVcc ){
		printk(KERN_DEBUG "Power, already set to %d\n", NewVcc);
		return 0;
	}

	ulSwitchSettings = EE_ADDRESS | ENABLE;
	switch( NewVcc )
	{
		case 0:
			printk(KERN_DEBUG "Configure the socket for 0 Volts\n");
			ulSwitchSettings |= AVCC_0V;
			break;

		case 50:
			printk(KERN_ERR "%s(): unsupported Vcc 5 Volts\n",
			       __FUNCTION__);
			return -1;

		case 33:
			printk(KERN_DEBUG "Configure the socket for 3.3 Volts\n");
			ulSwitchSettings |= AVCC_33V;
			break;

		default: 
			printk(KERN_ERR "%s(): unrecognized Vcc %u\n",
			       __FUNCTION__, Vcc);
			return -1;
	}

	outl((inl(GPIO_PADR) & ~0x8) |
	     ((ulSwitchSettings & AVCC_33V) ? 0x0 : 0x8), GPIO_PADR);
	mdelay(300);

	Vcc = NewVcc;

	printk(KERN_DEBUG  "ep93xx_set_voltage - exit\n");

	return 0;
}

static int armadillo9_pcmcia_configure_socket(struct soc_pcmcia_socket *skt,
				       const socket_state_t *state)
{
	int ret = 0;

	ep93xx_set_voltage(state->Vcc);

	if (!ret && skt->nr == 0) {
		if (state->flags & SS_RESET) {
			/* see EP93xx User's Guide */
			writel(0x15, SMC_PCMCIACtrl);
		} else {
			/* CF mode -  BIT1 for PCMCIA mode */
			writel(0x11, SMC_PCMCIACtrl);
		}
	}

	return ret;
}

static void armadillo9_pcmcia_socket_init(struct soc_pcmcia_socket *skt)
{
}

static void armadillo9_pcmcia_socket_suspend(struct soc_pcmcia_socket *skt)
{
}

static struct pcmcia_low_level armadillo9_pcmcia_ops = {
	.owner			= THIS_MODULE,
	.hw_init		= armadillo9_pcmcia_hw_init,
	.hw_shutdown		= armadillo9_pcmcia_hw_shutdown,
	.socket_state		= armadillo9_pcmcia_socket_state,
	.configure_socket	= armadillo9_pcmcia_configure_socket,
	.socket_init		= armadillo9_pcmcia_socket_init,
	.socket_suspend		= armadillo9_pcmcia_socket_suspend,
	.nr			= 1,
};

static struct platform_device *armadillo9_pcmcia_device;

static int __init armadillo9_pcmcia_init(void)
{
	int ret;

	armadillo9_pcmcia_device = kmalloc(sizeof(*armadillo9_pcmcia_device), GFP_KERNEL);
	if (!armadillo9_pcmcia_device)
		return -ENOMEM;
	memset(armadillo9_pcmcia_device, 0, sizeof(*armadillo9_pcmcia_device));
	armadillo9_pcmcia_device->name = "ep93xx-pcmcia";
	armadillo9_pcmcia_device->dev.platform_data = &armadillo9_pcmcia_ops;

	ret = platform_device_register(armadillo9_pcmcia_device);
	if (ret)
		kfree(armadillo9_pcmcia_device);

	return ret;
}

static void __exit armadillo9_pcmcia_exit(void)
{
	/*
	 * This call is supposed to free our armadillo9_pcmcia_device.
	 * Unfortunately platform_device don't have a free method, and
	 * we can't assume it's free of any reference at this point so we
	 * can't free it either.
	 */
	platform_device_unregister(armadillo9_pcmcia_device);
}

module_init(armadillo9_pcmcia_init);
module_exit(armadillo9_pcmcia_exit);

MODULE_LICENSE("GPL");
